#!/usr/bin/env python

# if you want to use me as a python module, please rename me to modunits.py
# if you want to use me as a standalone executable, please rename me to modunits
# if you want to use me as both, please name me whatever and symlink :)

import re
import sys
import string

def builtins():
	l = [ 'computer-size', 'computer-bit-seconds', 'computer-byte-hours', 'time' ]
	l.sort()
	return l

def detail_options():
	l = [ 'list', 'string-list', 'highest', 'two-highest', 'highest-and-fraction', 'all' ]
	l.sort()
	return l

def units_type():
	l = [ 'unabbreviated', 'abbreviated' ]
	l.sort()
	return l

#
# This one is very primitive class usage - it's basically just a struct, IE, no methods to speak of.  Using tuples and lists was getting too messy - this is
# much more descriptive
class typeclass:

	def __init__(self, threshold, unabbreviated, abbreviated):
		self.threshold = threshold
		self.unabbreviated = unabbreviated
		self.abbreviated = abbreviated


class itemclass:

	def __init__(self, amount, tc, after, units, number_of_digits_after_decimal_point):
		self.amount = amount
		self.tc = tc
		self.after = after
		self.units = units
		self.number_of_digits_after_decimal_point = number_of_digits_after_decimal_point

	# this handles units and before/after, but not things like commas or the detail level.  Comma processing and detail level processing will build on this
	def __str__(self):
		s = str(self.amount)
		if '.' in s and self.number_of_digits_after_decimal_point >= 1:
			fields = string.splitfields(s, '.')
			if len(fields) == 2:
				s = '%s.%s' % (fields[0], fields[1][0:self.number_of_digits_after_decimal_point])
		if self.units == 'unabbreviated':
			u = self.tc.unabbreviated
		elif self.units == 'abbreviated':
			u = self.tc.abbreviated
		else:
			sys.stderr.write('%s: Internal error: Illegal value for units\n' % sys.argv[0])
			usage(1)
		if self.after:
			if self.units == 'abbreviated':
				s += u
			else:
				s += ' '+u
		else:
			s = u+': '+s
		return s

def modunits(unitsvector, number, detail='highest-and-fraction', units="abbreviated", comma=1, reverse=0, after=1, number_of_digits_after_decimal_point=-1):

	iec=2**10
	si=10**3

	# detail can be:
	# 		list (implies that we will ignore units and comma)
	# 		highest (implies that we will ignore comma)
	# 		two highest
	# 		highest-and-fraction (default)
	# 		all
	# units can be:
	# 		unabbreviated
	#		abbreviated (default)
	# comma can be:
	#		0 (default)
	# 		1
	# reverse can be:
	#		0 (default)
	# 		1
	# after can be:
	#		0 (default)
	# 		1
	internalvector=unitsvector
	if type(internalvector) == type(''):
		if internalvector == 'time':
			# expects seconds
			internalnumber = number
			internalvector = [ \
				typeclass(60,'seconds','s'), \
				typeclass(60,'minutes','m'), \
				typeclass(24,'hours','h'), \
				typeclass(365, 'days','da'), \
				typeclass(100, 'years','y'), \
				typeclass(None, 'centuries','c'), \
				]
		elif internalvector in [ 'computer-size-iec' ]:
			# expects bytes
			internalnumber = number
			internalvector = [ \
				typeclass(iec,'bytes','B'), \
				typeclass(iec,'kibibytes','Ki'), \
				typeclass(iec,'mebibytes','Mi'), \
				typeclass(iec,'gibibytes','Gi'), \
				typeclass(iec,'tebibytes','Ti'), \
				typeclass(iec,'pebibytes','Pi'), \
				typeclass(iec,'exabytes','Ei'), \
				typeclass(iec,'zettabytes','Zi'), \
				typeclass(None,'yottabytes','Yi'), \
				]
		elif internalvector in [ 'computer-size', 'computer-size-si' ]:
			# expects bytes
			internalnumber = number
			internalvector = [ \
				typeclass(si,'bytes','B'), \
				typeclass(si,'kilobytes','k'), \
				typeclass(si,'megabytes','M'), \
				typeclass(si,'gigabytes','G'), \
				typeclass(si,'terabytes','T'), \
				typeclass(si,'petabytes','P'), \
				typeclass(si,'exabytes','E'), \
				typeclass(si,'zettabytes','Z'), \
				typeclass(None,'yottabytes','Y'), \
				]
		elif internalvector in [ 'computer-bit-seconds', 'computer-bits-per-second-iec' ]:
			# expects bits/second
			internalnumber = number
			internalvector = [ \
				typeclass(iec,'bits/s','b/s'), \
				typeclass(iec,'kibibits/second','ki/s'), \
				typeclass(iec,'mebibits/second','mi/s'), \
				typeclass(iec,'gibibits/second','gi/s'), \
				typeclass(iec,'tebibits/second','ti/s'), \
				typeclass(iec,'pebibits/second','pi/s'), \
				typeclass(iec,'exabits/second','ei/s'), \
				typeclass(iec,'zettabits/second','zi/s'), \
				typeclass(None,'yottabits/second','yi/s'), \
				]
		elif internalvector in [ 'computer-bits-per-second-si' ]:
			# expects bits/second
			internalnumber = number
			internalvector = [ \
				typeclass(si,'bits/s','b/s'), \
				typeclass(si,'kilobits/second','k/s'), \
				typeclass(si,'megabits/second','m/s'), \
				typeclass(si,'gigabits/second','g/s'), \
				typeclass(si,'terabits/second','t/s'), \
				typeclass(si,'petabits/second','p/s'), \
				typeclass(si,'exabits/second','e/s'), \
				typeclass(si,'zettabits/second','z/s'), \
				typeclass(None,'yottabits/second','y/s'), \
				]
		elif internalvector in [ 'computer-byte-hours', 'computer-bytes-per-hour-iec' ]:
			# expects bytes/hour
			internalnumber = number
			internalvector = [
				typeclass(iec,'bytes/hour','B/hr'), \
				typeclass(iec,'kibibytes/hour','Ki/hr'), \
				typeclass(iec,'mebibytes/hour','Mi/hr'), \
				typeclass(iec,'gibibytes/hour','Gi/hr'), \
				typeclass(iec,'tebibytes/hour','Ti/hr'), \
				typeclass(iec,'pebibytes/hour','Pi/hr'), \
				typeclass(iec,'exabytes/hour','Ei/hr'), \
				typeclass(iec,'zettabytes/hour','Zi/hr'), \
				typeclass(None,'yottabytes/hour','Yi/hr'), \
				]
		elif internalvector in [ 'computer-bytes-per-hour-si' ]:
			# expects bytes/hour
			internalnumber = number
			internalvector = [
				typeclass(si,'bytes/hour','B/hr'), \
				typeclass(si,'kilobytes/hour','k/hr'), \
				typeclass(si,'megabytes/hour','M/hr'), \
				typeclass(si,'gigabytes/hour','G/hr'), \
				typeclass(si,'terabytes/hour','T/hr'), \
				typeclass(si,'petabytes/hour','P/hr'), \
				typeclass(si,'exabytes/hour','E/hr'), \
				typeclass(si,'zettabytes/hour','Z/hr'), \
				typeclass(None,'yottabytes/hour','Y/hr'), \
				]
		else:
			sys.stderr.write('Sorry, I do not have a predefined units vector called %s\n' % unitsvector)
			sys.exit(1)
		return _chopit(internalvector, internalnumber, detail, units, comma, reverse, after, number_of_digits_after_decimal_point)
	else:
		# this may be broken since the switch to classes away from lists/tuples
		return _chopit(unitsvector, internalnumber, detail, units, comma, reverse, after, number_of_digits_after_decimal_point)

def _chopit(list, number, detail, units, comma, reverse, after, number_of_digits_after_decimal_point):
#	print '2 detail level is',detail
	# compute our list of numbers
	temp = number
	remainder = 0
	item_list = []
	for tc in list:
		if temp == 0:
			break
		if tc.threshold == None:
			(quotient, remainder) = 0, temp
		else:
			(quotient, remainder) = divmod(temp, tc.threshold)
			#print 'temp: %f, tc.threshold: %f, quotient: %f, remainder: %f' % (temp, tc.threshold, quotient, remainder)
		item_list.append(itemclass(long(remainder), tc, after, units, number_of_digits_after_decimal_point))
		if tc.threshold == None:
			break
#		print temp, quotient, remainder, item_list
		temp = quotient

	# handle the "0" case a little bit specially here, to avoid special cases below
	if len(item_list) == 0:
		item_list.append(itemclass(0, list[0], after, units, number_of_digits_after_decimal_point))

	# at this point, item_list should be a list of parts of the number, from least significant to most
	if detail == 'string-list':
		# pretty simple, and probably the easiest to build on :)  Just returns the list we built
		return map(str,item_list)
		# return early - don't fall through
	elif detail == 'list':
		# pretty simple, and probably the easiest to build on :)  Just returns the list we built
		return item_list
		# return early - don't fall through
	elif detail == 'highest':
		# chop off all but the most significant term - least significant are in the low-numbered list elements
		while item_list[1:]:
			del item_list[0]
		# and fall through
	elif detail == 'two-highest':
		# chop off all but the most significant two terms - least significant are in the low-numbered list elements
		while item_list[2:]:
			del item_list[0]
		# and fall through
	elif detail == 'highest-and-fraction':
		# this one's pretty different from the others, so we handle it here, entirely, rather than falling through for further processing.
		# we create a single "item" (of type item_class) that will hold the whole number + fraction for the most and 2nd most signficant items.  Then we return it
		# immediately.
		ic=item_list[-1]
		if item_list[1:]:
			# if we have at least two elements, then use the second from the end for the fraction
			ic.amount += item_list[-2].amount / float(item_list[-2].tc.threshold)
		return str(ic)
		# return early - don't fall through
	elif detail == 'all':
		# don't need to do anything for this one yet :)
		pass
		# and fall through
	else:
		sys.stderr.write('%s: Internal error: Illegal detail level %s in modunits.py\n' % (sys.argv[0], detail))

#	print '3 detail level is',detail
	# note that even detail == 'highest' works with this code
	if reverse:
		# and the units move right alone with the numbers
		item_list.reverse()

	if comma:
		separator=', '
	else:
		separator=' '

	return string.joinfields(map(str,item_list), separator)


def usage(retval):
	sys.stderr.write('Usage: %s [-h] -t type [-d detail] [-u unitstype] [-c] [-r] [-a] [-s] [-n number] [-D digits]\n')
	sys.stderr.write('-h\t\tgive this help message\n')
	sys.stderr.write('-t type\t\tSpecify the type of numbers to be reformatted.  Required.\n')
	sys.stderr.write('\t\tCurrently legal values are:\n')
	for Type in builtins():
		sys.stderr.write('\t\t\t%s\n' % Type)
	sys.stderr.write('-d detail\tSpecify the type of numbers to be reformatted\n')
	sys.stderr.write('\t\tCurrently legal values are:\n')
	for detail in detail_options():
		sys.stderr.write('\t\t\t%s\n' % detail)
	sys.stderr.write('-u unitstype\tSpecify if units should be abbreviated or spelled out in full\n')
	sys.stderr.write('\t\tCurrently legal values are:\n')
	for units in units_type():
		sys.stderr.write('\t\t\t%s\n' % units)
	sys.stderr.write('-c\t\tseparate with commas\n')
	sys.stderr.write('-r\t\toutput in reverse\n')
	sys.stderr.write('-a\t\tput the units after the numbers, not before.  A bit shorter\n')
	sys.stderr.write('-n number\tSpecify a number to be reformatted.  Can be repeated\n')
	sys.stderr.write('-s\t\tRead numbers from stdin\n')
	sys.stderr.write('\n')
	sys.stderr.write('-D\t\tNumber of digits to output after a decimal point (ignored for all but -d highest-and-fraction)\n')
	sys.stderr.write('\n')
	sys.stderr.write('You can specify both -s and -n, but if you do, the -n numbers\n')
	sys.stderr.write('\twill be processed before the numbers read from stdin.\n')
	sys.exit(retval)

def main():
	from_stdin=0
	from_argv=0
	numbers=[]
	units=''
	detail='highest-and-fraction'
	units="abbreviated"
	comma=0
	reverse=0
	after=0
	Type = ''
	number_of_digits_after_decimal_point = -1
	while sys.argv[1:]:
		if sys.argv[1] in [ '-h', '--help' ]:
			usage(0)
		elif sys.argv[1] == '-t' and sys.argv[2:]:
			Type=sys.argv[2]
			del sys.argv[1]
		elif sys.argv[1] == '-d' and sys.argv[2:]:
			if sys.argv[2] in detail_options():
				detail=sys.argv[2]
			else:
				sys.stderr.write('%s: Illegal detail spec %s\n' % (sys.argv[0], sys.argv[2]))
				usage(1)
			del sys.argv[1]
		elif sys.argv[1] == '-u' and sys.argv[2:]:
			if sys.argv[2] in units_type():
				units=sys.argv[2]
			else:
				sys.stderr.write('%s: Illegal units spec %s\n' % (sys.argv[0], sys.argv[2]))
				usage(1)
			del sys.argv[1]
		elif sys.argv[1] == '-c':
			comma=1
		elif sys.argv[1] == '-r':
			reverse=1
		elif sys.argv[1] == '-a':
			after=1
		elif sys.argv[1] == '-s':
			from_stdin=1
		elif sys.argv[1] == '-n' and sys.argv[2:]:
			from_argv=1
			whole_number_string = re.sub('\.[0-9]*$', '', sys.argv[2])
			if long(whole_number_string) == float(sys.argv[2]):
				numbers.append(long(whole_number_string))
			else:
				numbers.append(float(string.strip(sys.argv[2])))
			del sys.argv[1]
		elif sys.argv[1] == '-D' and sys.argv[2:]:
			number_of_digits_after_decimal_point = string.atoi(string.strip(sys.argv[2]))
			del sys.argv[1]
		else:
			sys.stderr.write('%s: Illegal option %s\n' % (sys.argv[0], sys.argv[1]))
			usage(1)
		del sys.argv[1]

#	print '1 detail level is',detail
	if Type=='':
		sys.stderr.write('%s: -u is a required option\n' % sys.argv[0])
		usage(1)
		
	if from_argv:
		for number in numbers:
			print modunits(Type, int(number), detail, units, comma, reverse, after, number_of_digits_after_decimal_point)

	if from_stdin:
		while 1:
			line = sys.stdin.readline()
			if not line:
				break
			print modunits(Type, int(string.atof(string.strip(line))), detail, units, comma, reverse, after, number_of_digits_after_decimal_point)

	if not from_argv and not from_stdin:
		sys.stderr.write('%s: Warning: You probably want to specify either -s -or -n\n' % sys.argv[0])

	sys.exit(0)
	
if __name__ == '__main__':
    main()

